def test_nlargest(Chain):
    assert Chain([]).nlargest(0).value == []
    assert Chain([]).nlargest(1).value == []
    assert Chain([1, 3, 2, 4, 3]).nlargest(2).value == [4, 3]
    assert Chain([1, 3, 2, 4, 3]).nlargest(3).value == [4, 3, 3]
    assert Chain([1, 3, 2, 4, 1]).nlargest(0).value == []
    assert Chain([1, 3, 2, 4, 3]).nlargest(1, key=lambda x: x % 3).value == [2]


def test_nsmallest(Chain):
    assert Chain([]).nsmallest(0).value == []
    assert Chain([]).nsmallest(1).value == []
    assert Chain([1, 3, 2, 4, 1]).nsmallest(2).value == [1, 1]
    assert Chain([1, 3, 2, 4, 1]).nsmallest(3).value == [1, 1, 2]
    assert Chain([1, 3, 2, 4, 1]).nsmallest(0).value == []
    assert Chain([1, 3, 2, 4, 3]).nsmallest(2, key=lambda x: x % 3).value == [3, 3]



import heapq

from funcy_chain import Chain


def test_empty_chain():
    assert Chain([]) != []
    assert Chain([]).value == []


def test_tap():
    arr = []
    assert Chain([1, 2, 3]).tap(arr.append).value == [1, 2, 3]
    assert arr == [[1, 2, 3]]


def test_thru():
    assert Chain([1, 3, 2]).thru(sorted).value == [1, 2, 3]
    assert Chain([1, 3, 2]).thru(lambda arr: heapq.nlargest(2, arr)).value == [3, 2]



import sys
from test.conftest import IterChainWrapper

import pytest


def test_enumerate(Chain):
    assert Chain([]).enumerate().value == []
    assert Chain([1, 2, 3]).enumerate().value == [(0, 1), (1, 2), (2, 3)]


def test_max(Chain):
    assert Chain([1, 3, 2]).max().value == 3
    assert Chain([1, 3, 2]).max(key=lambda x: -x).value == 1

    with pytest.raises(ValueError):
        Chain([]).max()


def test_min(Chain):
    assert Chain([1, 3, 2]).min().value == 1
    assert Chain([1, 3, 2]).min(key=lambda x: -x).value == 3

    with pytest.raises(ValueError):
        Chain([]).min()


def test_reduce(Chain):
    add = lambda a, b: a + b

    # empty lists
    assert Chain([]).reduce(add, 0).value == 0
    assert Chain([]).reduce(add, []).value == []
    with pytest.raises(Exception):
        Chain([]).reduce(add).value

    # adding ints
    assert Chain([1, 2, 3]).reduce(add).value == 6
    assert Chain([1, 2, 3]).reduce(add, 0).value == 6
    assert Chain([1, 2, 3]).reduce(add, 4).value == 10

    # list concatenation
    assert Chain([[1], [2]]).reduce(add).value == [1, 2]
    assert Chain([[1], [2]]).reduce(add, []).value == [1, 2]
    with pytest.raises(TypeError):
        Chain([[1], [2]]).reduce(add, 0).value


def test_reverse(Chain):
    assert Chain([]).reverse().value == []
    assert Chain([1, 2, 3]).reverse().value == [3, 2, 1]
    assert Chain([1, 2, 3]).reverse().reverse().value == [1, 2, 3]
    assert Chain([[1], [2]]).reverse().value == [[2], [1]]


def test_slice(Chain):
    assert Chain([]).slice(1).value == []
    assert Chain([1, 2, 3]).slice(3).value == [1, 2, 3]
    assert Chain([1, 2, 3]).slice(0, 3).value == [1, 2, 3]
    assert Chain([1, 2, 3]).slice(2).value == [1, 2]
    assert Chain([1, 2, 3]).slice(0, 2).value == [1, 2]
    assert Chain([1, 2, 3]).slice(1, 3).value == [2, 3]
    assert Chain([1, 2, 3]).slice(1, None).value == [2, 3]
    assert Chain([1, 2, 3]).slice(2, None).value == [3]
    assert Chain([1, 2, 3]).slice(3, None).value == []
    assert Chain([1, 2, 3]).slice(4, None).value == []
    assert Chain([1, 2, 3]).slice(1, 2).value == [2]
    assert Chain([1, 2, 3]).slice(0, 3).value == [1, 2, 3]

    assert Chain([1, 2, 3, 4, 5, 6, 7, 8]).slice(0, 8, 2).value == [1, 3, 5, 7]

    if not issubclass(Chain, IterChainWrapper):
        assert Chain([1, 2, 3]).slice(0, 3, -1).value == []
        assert Chain([1, 2, 3]).slice(3, 0, -1).value == [3, 2]
        assert Chain([1, 2, 3]).slice(3, None, -1).value == [3, 2, 1]
        assert Chain([1, 2, 3]).slice(0, -1).value == [1, 2]
        assert Chain([1, 2, 3]).slice(1, -1).value == [2]
        assert Chain([1, 2, 3]).slice(-1, 3).value == [3]
        assert Chain([1, 2, 3]).slice(-1, 1).value == []
        assert Chain([1, 2, 3, 4, 5, 6, 7, 8]).slice(8, None, -2).value == [8, 6, 4, 2]
        assert Chain([1, 2, 3, 4, 5, 6, 7, 8]).slice(8, 2, -2).value == [8, 6, 4]


def test_sort(Chain):
    assert Chain([]).sort().value == []
    assert Chain([1, 3, 2]).sort().value == [1, 2, 3]
    assert Chain([1, 3, 2]).sort(reverse=True).value == [3, 2, 1]
    assert Chain([1, 3, 2]).sort(key=lambda x: -x).value == [3, 2, 1]
    assert Chain([1, 3, 2]).sort(key=lambda x: -x, reverse=True).value == [1, 2, 3]
    assert Chain([1, 3, 2]).sort(lambda x: -x, True).value == [1, 2, 3]


def test_sum(Chain):
    assert Chain([]).sum().value == 0
    assert Chain([1, 3, 2]).sum().value == 6
    assert Chain([1, 3, 2]).sum(4).value == 10
    assert Chain([1, 3, 2]).sum(start=4).value == 10

    with pytest.raises(TypeError):
        Chain([[1], [2]]).sum().value
    assert Chain([[1], [2]]).sum([]).value == [1, 2]
    assert Chain([[1], [2]]).sum(start=[]).value == [1, 2]


def test_zip(Chain):
    assert Chain([]).zip().value == []
    assert Chain([()]).zip().value == []
    assert Chain([(), ()]).zip().value == []
    assert Chain([(1,), (3,)]).zip().value == [(1, 3)]
    assert Chain([(1, 2)]).zip().value == [(1,), (2,)]
    assert Chain([(1, 2), ()]).zip().value == []
    assert Chain([(1, 2), (3,)]).zip().value == [(1, 3)]
    assert Chain([(1,), (3, 4)]).zip().value == [(1, 3)]
    assert Chain([(1, 2), (3, 4)]).zip().value == [(1, 3), (2, 4)]

    if sys.version_info >= (3, 10):
        with pytest.raises(ValueError):
            Chain([(1, 2), ()]).zip(strict=True).value
        with pytest.raises(ValueError):
            Chain([(1, 2), (3,)]).zip(strict=True).value
        with pytest.raises(ValueError):
            Chain([(1,), (3, 4)]).zip(strict=True).value



def test_names(Chain):
    data = {
        "user1": {
            "firstname": "Alice",
            "lastname": "Liddle",
        },
        "user2": {
            "firstname": "Bob",
            "lastname": "Kennedy",
        },
    }
    names = (
        Chain(data)
        .values()
        .map(lambda user_data: "{lastname}, {firstname}".format(**user_data))
        .sort()
    ).value
    assert names == ["Kennedy, Bob", "Liddle, Alice"]


def test_integers(Chain):
    assert (
        Chain([1, 2, 3, 7, 6, 5, 4])
        .without(3)
        .filter(lambda x: x > 2)
        .remove(lambda x: x > 6)
        .sort(reverse=True)
    ).value == [6, 5, 4]


def test_youngest(Chain):
    users = [
        {"username": "barney", "age": 36},
        {"username": "fred", "age": 40},
        {"username": "pebbles", "age": 1},
    ]
    names = (Chain(users).sort(lambda user: user["age"]).map(lambda user: user["username"])).value
    assert names == ["pebbles", "barney", "fred"]



import sys
import warnings

import pytest


def test_choice(Chain):
    assert Chain([1, 2, 3]).choice().value in [1, 2, 3]
    with pytest.raises(IndexError):
        Chain([]).choice().value


def test_choices(Chain):
    assert Chain([1, 2, 3]).choices(k=1).value in [[1], [2], [3]]
    assert len(Chain([1, 2, 3]).choices(k=4).value) == 4
    assert len(Chain([1, 2, 3]).choices([1, 2, 3], k=4).value) == 4
    assert len(Chain([1, 2, 3]).choices(weights=[1, 2, 3], k=4).value) == 4
    assert len(Chain([1, 2, 3]).choices(cum_weights=[1, 2, 3], k=4).value) == 4
    with pytest.raises(IndexError):
        Chain([]).choices(k=1).value


def test_sample(Chain):
    assert Chain([1, 2, 3]).sample(1).value in [[1], [2], [3]]
    assert len(Chain([1, 2, 3]).sample(2).value) == 2
    assert len(Chain([1, 2, 3]).sample(3).value) == 3
    with pytest.raises(ValueError):
        Chain([]).sample(1).value
    with pytest.raises(ValueError):
        Chain([1, 2, 3]).sample(4).value

    if sys.version_info >= (3, 9):
        assert len(Chain([1, 2, 3]).sample(8, counts=[3, 3, 3]).value) == 8
        assert set(Chain([1, 2, 3]).sample(8, counts=[3, 3, 3]).value) <= {1, 2, 3}
        with pytest.raises(ValueError):
            Chain([1, 2, 3]).sample(10, counts=[3, 3, 3]).value


def test_shuffle(Chain):
    assert Chain([]).shuffle().value == []
    assert Chain([1, 2, 3]).shuffle().sort().value == [1, 2, 3]


# The "random" argument to random.shuffle() was removed in Python 3.11.
@pytest.mark.skipif(sys.version_info >= (3, 11), reason="'random' argument remove in Python 3.11")
def test_shuffle_with_random_argument(Chain):
    # The "random" argument for random.shuffle() is deprecated since Python 3.9.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert Chain([1, 2, 3]).shuffle(random=lambda: 0.5).sort().value == [1, 2, 3]



from test.conftest import IterChainWrapper

from funcy_chain import IterChain


def test_drop(Chain):
    assert Chain([]).drop(0).value == []
    assert Chain([]).drop(1).value == []
    assert Chain([1, 2]).drop(0).value == [1, 2]
    assert Chain([1, 2]).drop(1).value == [2]
    assert Chain([1, 2]).drop(2).value == []
    assert Chain([1, 2]).drop(3).value == []


def test_take(Chain):
    assert Chain([]).take(0).value == []
    assert Chain([]).take(1).value == []
    assert Chain([1, 2]).take(0).value == []
    assert Chain([1, 2]).take(1).value == [1]
    assert Chain([1, 2]).take(2).value == [1, 2]
    assert Chain([1, 2]).take(3).value == [1, 2]


def test_rest(Chain):
    assert Chain([]).rest().value == []
    assert Chain([1]).rest().value == []
    assert Chain([1, 2]).rest().value == [2]


def test_butlast(Chain):
    assert Chain([]).butlast().value == []
    assert Chain([1]).butlast().value == []
    assert Chain([1, 2]).butlast().value == [1]


def test_map(Chain):
    assert Chain([]).map(lambda x: x.foo).value == []
    assert Chain([]).map(lambda x: x + 1).value == []
    assert Chain([1, 2]).map(lambda x: x + 1).value == [2, 3]


def test_filter(Chain):
    assert Chain([]).filter(lambda x: x.foo).value == []
    assert Chain([]).filter(lambda x: x > 1).value == []
    assert Chain([1, 2]).filter(lambda x: x > 1).value == [2]


def test_remove(Chain):
    assert Chain([]).remove(lambda x: x).value == []
    assert Chain([0, 1, 2]).remove(lambda x: x != 0).value == [0]
    assert Chain([0, 1, 2]).remove(lambda x: x == 1).value == [0, 2]


def test_keep(Chain):
    assert Chain([]).keep().value == []
    assert Chain([]).keep(lambda x: x).value == []
    assert Chain([0, 1, 2]).keep().value == [1, 2]
    assert Chain([0, 1, 2]).keep(lambda x: x - 1).value == [-1, 1]


def test_without(Chain):
    assert Chain([]).without(2).value == []
    assert Chain([1, 2, 3]).without().value == [1, 2, 3]
    assert Chain([1, 2, 3]).without(3).value == [1, 2]
    assert Chain([1, 2, 3]).without(2).value == [1, 3]
    assert Chain([1, 2, 3]).without(2, 3).value == [1]
    assert Chain([1, 2, 3]).without(3, 2, 3).value == [1]
    assert Chain([1, 2, 3]).without(1, 2, 3).value == []


def test_concat(Chain):
    assert Chain([]).concat().value == []
    assert Chain([[]]).concat().value == []
    assert Chain([[], []]).concat().value == []
    assert Chain([[1, 2]]).concat().value == [1, 2]
    assert Chain([[1, 2], [3]]).concat().value == [1, 2, 3]


def test_flatten(Chain):
    assert Chain([]).flatten().value == []
    assert Chain([[]]).flatten().value == []
    assert Chain([[], []]).flatten().value == []
    assert Chain([1, [2], 3]).flatten().value == [1, 2, 3]


def test_mapcat(Chain):
    assert Chain([]).mapcat(lambda x: x.foo).value == []
    assert Chain([]).mapcat(lambda x: [x + 1]).value == []
    assert Chain([1, 2]).mapcat(lambda x: [x + 1]).value == [2, 3]
    assert Chain([1, 2]).mapcat(lambda x: [x, -x]).value == [1, -1, 2, -2]


def test_interleave(Chain):
    assert Chain([]).interleave().value == []
    assert Chain([[1]]).interleave().value == [1]
    assert Chain([[1, 2]]).interleave().value == [1, 2]
    assert Chain([[1], [2]]).interleave().value == [1, 2]
    assert Chain([[1, 2], [3, 4]]).interleave().value == [1, 3, 2, 4]


def test_interpose(Chain):
    assert Chain([]).interpose("X").value == []
    assert Chain(["a"]).interpose("X").value == ["a"]
    assert Chain(["a", "b"]).interpose("X").value == ["a", "X", "b"]
    assert Chain(["a", "b", "c"]).interpose("X").value == ["a", "X", "b", "X", "c"]


def test_takewhile(Chain):
    assert Chain([]).takewhile(lambda x: x.foo).value == []
    assert Chain([-1, 0, 1, 2, 3, 4]).takewhile().value == [-1]
    assert Chain([-1, 0, 1, 2, 3, 4]).takewhile(lambda x: x != 3).value == [-1, 0, 1, 2]


def test_dropwhile(Chain):
    assert Chain([]).dropwhile(lambda x: x.foo).value == []
    assert Chain([-1, 0, 1, 2, 3, 4]).dropwhile().value == [0, 1, 2, 3, 4]
    assert Chain([-1, 0, 1, 2, 3, 4]).dropwhile(lambda x: x != 3).value == [3, 4]


def test_distinct(Chain):
    assert Chain([]).distinct().value == []
    assert Chain([1]).distinct().value == [1]
    assert Chain([1, 2, 1, 3]).distinct().value == [1, 2, 3]


def test_split(Chain):
    if issubclass(Chain, IterChainWrapper):
        Chain = IterChain

    def unlazy(chain_value):
        return tuple(map(list, chain_value))

    assert unlazy(Chain([]).split(lambda x: x.foo).value) == ([], [])
    assert unlazy(Chain([]).split(lambda x: x > 0).value) == ([], [])
    assert unlazy(Chain([-1, 1, 0]).split(lambda x: x > 0).value) == ([1], [-1, 0])


def test_split_at(Chain):
    if issubclass(Chain, IterChainWrapper):
        Chain = IterChain

    def unlazy(chain_value):
        return tuple(map(list, chain_value))

    assert unlazy(Chain([]).split_at(0).value) == ([], [])
    assert unlazy(Chain(["a", "b", "c"]).split_at(0).value) == ([], ["a", "b", "c"])
    assert unlazy(Chain(["a", "b", "c"]).split_at(1).value) == (["a"], ["b", "c"])
    assert unlazy(Chain(["a", "b", "c"]).split_at(3).value) == (["a", "b", "c"], [])
    assert unlazy(Chain(["a", "b", "c"]).split_at(4).value) == (["a", "b", "c"], [])


def test_split_by(Chain):
    if issubclass(Chain, IterChainWrapper):
        Chain = IterChain

    def unlazy(chain_value):
        return tuple(map(list, chain_value))

    assert unlazy(Chain([]).split_by(lambda x: x.foo).value) == ([], [])
    assert unlazy(Chain([]).split_by(lambda x: x > 0).value) == ([], [])
    assert unlazy(Chain([-1, 1, -1]).split_by(lambda x: x < 0).value) == ([-1], [1, -1])


def test_group_by(Chain):
    assert Chain([]).group_by(lambda x: x.foo).value == {}
    assert Chain([1, 2, 3]).group_by(lambda x: x % 2).value == {0: [2], 1: [1, 3]}


def test_group_by_keys(Chain):
    assert Chain([]).group_by_keys(lambda x: x.foo).value == {}
    assert Chain(["Tom", "Tim"]).group_by_keys(lambda x: (x[0], x[1])).value == {
        "T": ["Tom", "Tim"],
        "o": ["Tom"],
        "i": ["Tim"],
    }


def test_group_values(Chain):
    assert Chain([]).group_values().value == {}
    assert Chain([("a", 1), ("b", 2), ("a", 3)]).group_values().value == {"b": [2], "a": [1, 3]}


def test_count_by(Chain):
    assert Chain([]).count_by(lambda x: x.foo).value == {}
    assert Chain([1, 2, 3]).count_by(lambda x: x % 2).value == {1: 2, 0: 1}


def test_count_reps(Chain):
    assert Chain([]).count_reps().value == {}
    assert Chain([1, 2, 1]).count_reps().value == {1: 2, 2: 1}


def test_partition(Chain):
    assert Chain([]).partition(2).value == []
    assert Chain([1]).partition(2).value == []
    assert Chain([1, 2]).partition(2).value == [[1, 2]]
    assert Chain([1, 2, 3]).partition(2).value == [[1, 2]]
    assert Chain([1, 2, 3, 4, 5]).partition(2, 2).value == [[1, 2], [3, 4]]
    assert Chain([1, 2, 3, 4, 5]).partition(2, 3).value == [[1, 2], [4, 5]]
    assert Chain([1, 2, 3, 4, 5, 6, 7]).partition(2, 3).value == [[1, 2], [4, 5]]


def test_chunks(Chain):
    assert Chain([]).chunks(2).value == []
    assert Chain([1]).chunks(2).value == [[1]]
    assert Chain([1, 2]).chunks(2).value == [[1, 2]]
    assert Chain([1, 2, 3]).chunks(2).value == [[1, 2], [3]]
    assert Chain([1, 2, 3, 4, 5]).chunks(2, 2).value == [[1, 2], [3, 4], [5]]
    assert Chain([1, 2, 3, 4, 5]).chunks(2, 3).value == [[1, 2], [4, 5]]
    assert Chain([1, 2, 3, 4, 5, 6, 7]).chunks(2, 3).value == [[1, 2], [4, 5], [7]]


def test_partition_by(Chain):
    if issubclass(Chain, IterChainWrapper):
        Chain = IterChain

    def unlazy(chain_value):
        return list(map(list, chain_value))

    assert unlazy(Chain([]).partition_by(lambda x: x.foo).value) == []
    assert unlazy(Chain([]).partition_by(lambda x: x % 2).value) == []
    assert unlazy(Chain([1]).partition_by(lambda x: x % 2).value) == [[1]]
    assert unlazy(Chain([1, 2, 4, 3, 5, 2]).partition_by(lambda x: x % 2).value) == [
        [1],
        [2, 4],
        [3, 5],
        [2],
    ]


def test_with_prev(Chain):
    assert Chain([]).with_prev().value == []
    assert Chain([1]).with_prev().value == [(1, None)]
    assert Chain([1]).with_prev(0).value == [(1, 0)]
    assert Chain([1]).with_prev(fill=0).value == [(1, 0)]
    assert Chain([1, 2, 3]).with_prev().value == [(1, None), (2, 1), (3, 2)]


def test_with_next(Chain):
    assert Chain([]).with_next().value == []
    assert Chain([1]).with_next().value == [(1, None)]
    assert Chain([1]).with_next(0).value == [(1, 0)]
    assert Chain([1]).with_next(fill=0).value == [(1, 0)]
    assert Chain([1, 2, 3]).with_next().value == [(1, 2), (2, 3), (3, None)]


def test_pairwise(Chain):
    assert Chain([]).pairwise().value == []
    assert Chain([1]).pairwise().value == []
    assert Chain([1, 2, 3]).pairwise().value == [(1, 2), (2, 3)]


def test_accumulate(Chain):
    assert Chain([]).accumulate().value == []
    assert Chain([]).accumulate(lambda a, b: a.foo).value == []
    assert Chain([1, 2, 3, 4]).accumulate().value == [1, 3, 6, 10]
    assert Chain([1, 2, 3, 4]).accumulate(lambda acc, x: acc * x).value == [1, 2, 6, 24]


def test_reductions(Chain):
    add = lambda a, b: a + b
    mul = lambda a, b: a * b

    assert Chain([]).reductions(lambda a, b: a.foo).value == []
    assert Chain([]).reductions(add).value == []
    assert Chain([1, 2, 3, 4]).reductions(add).value == [1, 3, 6, 10]
    assert Chain([1, 2, 3, 4]).reductions(mul).value == [1, 2, 6, 24]
    assert Chain([1, 2, 3, 4]).reductions(add, acc=-1).value == [0, 2, 5, 9]


def test_sums(Chain):
    assert Chain([]).sums().value == []
    assert Chain([1, 2, 3, 4]).sums().value == [1, 3, 6, 10]
    assert Chain([1, 2, 3, 4]).sums(acc=-1).value == [0, 2, 5, 9]



#!/usr/bin/env python
import subprocess
import sys

if __name__ == "__main__":
    args = sys.argv[1:]
    if "--no-deps" in args:
        subprocess.run(["flit", "install", "--deps", "none"], check=True)
    else:
        subprocess.run(
            ["pip", "install", "--disable-pip-version-check", "--no-python-version-warning"] + args,
            check=True,
        )
        subprocess.run(["flit", "install", "--extras", "test"], check=True)



import pytest


def test_items(Chain):
    assert Chain({}).items().value == []
    assert Chain({1: 2}).items().value == [(1, 2)]

    with pytest.raises((AttributeError, TypeError)):
        assert Chain([]).items().value


def test_keys(Chain):
    assert Chain({}).keys().value == []
    assert Chain({1: 2}).keys().value == [1]

    with pytest.raises((AttributeError, TypeError)):
        assert Chain([]).keys().value


def test_values(Chain):
    assert Chain({}).values().value == []
    assert Chain({1: 2}).values().value == [2]

    with pytest.raises((AttributeError, TypeError)):
        assert Chain([]).values().value


def test_update(Chain):
    assert Chain({}).update().value == {}
    assert Chain({}).update({"b": 2}).value == {"b": 2}
    assert Chain({}).update([("b", 2)]).value == {"b": 2}
    assert Chain({}).update(b=2).value == {"b": 2}
    assert Chain({"a": 1}).update({"b": 2}).value == {"a": 1, "b": 2}
    assert Chain({"a": 1}).update({"a": 2}).value == {"a": 2}

    with pytest.raises((AttributeError, TypeError)):
        assert Chain([]).update().value



def test_starmap(Chain):
    assert Chain([]).starmap(lambda x: x.foo).value == []
    assert Chain([(1, 2), (2, 3), (3, 4)]).starmap(pow).value == [1, 8, 81]


def test_zip_longest(Chain):
    assert Chain([]).zip_longest().value == []
    assert Chain([()]).zip_longest().value == []
    assert Chain([(), ()]).zip_longest().value == []
    assert Chain([(1,), (3,)]).zip_longest().value == [(1, 3)]
    assert Chain([(1, 2)]).zip_longest().value == [(1,), (2,)]
    assert Chain([(1, 2), ()]).zip_longest().value == [(1, None), (2, None)]
    assert Chain([(1, 2), (3,)]).zip_longest().value == [(1, 3), (2, None)]
    assert Chain([(1,), (3, 4)]).zip_longest().value == [(1, 3), (None, 4)]
    assert Chain([(1, 2), (3,)]).zip_longest(fillvalue=0).value == [(1, 3), (2, 0)]
    assert Chain([(1, 2), (3, 4)]).zip_longest().value == [(1, 3), (2, 4)]
    assert Chain([(1, 2), (3, 4)]).zip_longest(fillvalue=0).value == [(1, 3), (2, 4)]



import abc

__all__ = [
    "ChainBase",
]


class ChainBase(metaclass=abc.ABCMeta):
    """Base class for method chaining implementations."""

    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def tap(self, interceptor):
        interceptor(self._value)
        return self

    def thru(self, interceptor):
        return self.__class__(interceptor(self._value))



from test.conftest import IterChainWrapper

import pytest

from funcy_chain import IterChain


def test_join(Chain):
    Chain([]).join().value  # Check that no exception is raised.
    assert Chain(["a", "b"]).join().value == "ab"
    assert Chain([[1], [2]]).join().value == [1, 2]
    assert Chain([(1,), (2,)]).join().value == (1, 2)
    assert Chain([{"a": 1, "b": 2}, {"b": -2, "c": 3}]).join().value == {"a": 1, "b": -2, "c": 3}
    assert Chain([{"a": 1, "b": 2}, {"b": -2, "c": 3}]).merge().value == {"a": 1, "b": -2, "c": 3}


def test_join_with(Chain):
    Chain([]).join_with(list).value  # Check that no exception is raised.
    assert Chain([{"a": 1, "b": 2}, {"b": -2, "c": 3}]).join_with(list).value == {
        "a": [1],
        "b": [2, -2],
        "c": [3],
    }
    assert Chain([{"a": 1, "b": 2}, {"b": -2, "c": 3}]).merge_with(list).value == {
        "a": [1],
        "b": [2, -2],
        "c": [3],
    }
    assert Chain([{"a": 1, "b": 2}, {"b": -2, "c": 3}]).join_with(tuple).value == {
        "a": (1,),
        "b": (2, -2),
        "c": (3,),
    }


def test_walk(Chain):
    assert Chain([1, 2]).walk(lambda x: x + 1).value == [2, 3]
    assert Chain((1, 2)).walk(lambda x: x + 1).value == (2, 3)


def test_walk_keys(Chain):
    assert Chain({1: 1, 2: 2}).walk_keys(lambda x: x + 1).value == {2: 1, 3: 2}


def test_walk_values(Chain):
    assert Chain({1: 1, 2: 2}).walk_values(lambda x: x + 1).value == {1: 2, 2: 3}


def test_select(Chain):
    assert Chain([]).select(lambda x: x > 1).value == []
    assert Chain([1, 2]).select(lambda x: x > 1).value == [2]
    if not issubclass(Chain, IterChainWrapper):
        assert Chain(()).select(lambda x: x > 1).value == ()
        assert Chain((1, 2)).select(lambda x: x > 1).value == (2,)
        assert Chain({1, 2}).select(lambda x: x > 1).value == {2}


def test_select_keys(Chain):
    assert Chain({}).select_keys(lambda x: x > 1).value == {}
    assert Chain({1: 1, 2: 2}).select_keys(lambda x: x > 1).value == {2: 2}


def test_select_values(Chain):
    assert Chain({}).select_values(lambda x: x > 1).value == {}
    assert Chain({1: 1, 2: 2}).select_values(lambda x: x > 1).value == {2: 2}


def test_compact(Chain):
    assert Chain([]).compact().value == []
    if not issubclass(Chain, IterChainWrapper):
        assert Chain(()).compact().value == ()
        assert Chain({}).compact().value == {}
    assert Chain([0, 2]).compact().value == [2]
    if not issubclass(Chain, IterChainWrapper):
        assert Chain((0, 2)).compact().value == (2,)
        assert Chain({0, 2}).compact().value == {2}
        assert Chain({0: 1, 1: 0, 2: 2}).compact().value == {0: 1, 2: 2}


def test_flip(Chain):
    assert Chain({}).flip().value == {}
    assert Chain({1: 2}).flip().value == {2: 1}


def test_project(Chain):
    assert Chain({}).project([]).value == {}
    assert Chain({}).project([1]).value == {}
    assert Chain({1: "a"}).project([1]).value == {1: "a"}
    assert Chain({2: "b"}).project([1]).value == {}
    assert Chain({1: "a", 2: "b"}).project([1]).value == {1: "a"}


def test_omit(Chain):
    assert Chain({}).omit([]).value == {}
    assert Chain({}).omit([1]).value == {}
    assert Chain({1: "a"}).omit([1]).value == {}
    assert Chain({2: "b"}).omit([1]).value == {2: "b"}
    assert Chain({1: "a", 2: "b"}).omit([1]).value == {2: "b"}


def test_zip_values(Chain):
    assert Chain([]).zip_values().value == []
    assert Chain([{}]).zip_values().value == []
    assert Chain([{1: "a"}]).zip_values().value == [("a",)]
    assert Chain([{1: "a"}, {1: "aa"}]).zip_values().value == [("a", "aa")]
    assert Chain([{1: "a"}, {2: "b"}]).zip_values().value == []
    assert Chain([{1: "a", 2: "b"}, {1: "aa"}]).zip_values().sort().value == [("a", "aa")]
    assert Chain([{1: "a", 2: "b"}, {2: "aa"}]).zip_values().sort().value == [("b", "aa")]


def test_zip_dicts(Chain):
    assert Chain([]).zip_dicts().value == []
    assert Chain([{}]).zip_dicts().value == []
    assert Chain([{1: "a"}]).zip_dicts().value == [(1, ("a",))]
    assert Chain([{1: "a"}, {1: "aa"}]).zip_dicts().value == [(1, ("a", "aa"))]
    assert Chain([{1: "a"}, {2: "b"}]).zip_dicts().value == []
    assert Chain([{1: "a", 2: "b"}, {1: "aa"}]).zip_dicts().sort().value == [(1, ("a", "aa"))]
    assert Chain([{1: "a", 2: "b"}, {2: "aa"}]).zip_dicts().sort().value == [(2, ("b", "aa"))]


def test_where(Chain):
    assert Chain([]).where(a=1).value == []
    assert Chain([{}]).where(a=1).value == []
    assert Chain([{"a": 1}]).where(a=1).value == [{"a": 1}]
    assert Chain([{"a": 2}]).where(a=1).value == []
    assert Chain([{"b": 1}]).where(a=1).value == []
    assert Chain([{"a": 1}, {"a": 2}]).where(a=1).value == [{"a": 1}]


def test_pluck(Chain):
    assert Chain([]).pluck("a").value == []
    with pytest.raises(KeyError):
        Chain([{}]).pluck("a").value
    assert Chain([{"a": 1}]).pluck("a").value == [1]
    with pytest.raises(KeyError):
        Chain([{"a": 1}]).pluck("b").value
    assert Chain([{"a": 1}, {"a": 2, "b": 3}]).pluck("a").value == [1, 2]


def test_pluck_attr(Chain):
    if issubclass(Chain, IterChainWrapper):
        Chain = IterChain
    assert list(Chain([]).pluck_attr("imag").value) == []
    with pytest.raises(AttributeError):
        list(Chain(["a"]).pluck_attr("imag").value)
    assert list(Chain([1j]).pluck_attr("imag").value) == [1]
    with pytest.raises(AttributeError):
        list(Chain([1j]).pluck_attr("foo").value)
    assert list(Chain([1j, 2j]).pluck_attr("imag").value) == [1, 2]


def test_invoke(Chain):
    assert Chain([]).invoke("foo").value == []
    assert Chain([1j]).invoke("conjugate").value == [-1j]
    assert Chain([[1, 2], [3, 4]]).invoke("count", 1).value == [1, 0]



import sys
from functools import reduce
from heapq import nlargest, nsmallest
from itertools import islice, starmap, zip_longest
from random import choice, choices, sample, shuffle

import funcy
from funcy.funcmakers import make_func, make_pred

from .chain_base import ChainBase
from .utils import UNSET


class IterChain(ChainBase):

    ## builtins

    def enumerate(self, start=0):
        return IterChain(enumerate(self._value, start=start))

    def max(self, key=None):
        return IterChain(max(self._value, key=None if key is None else make_func(key)))

    def min(self, key=None):
        return IterChain(min(self._value, key=None if key is None else make_func(key)))

    def reduce(self, f, *initializer):
        return IterChain(reduce(make_func(make_func(f)), self._value, *initializer))

    def reverse(self):
        try:
            value = reversed(self._value)
        except TypeError:
            value = reversed(list(self._value))
        return IterChain(value)

    def slice(self, *args):
        return IterChain(islice(self._value, *args))

    def sort(self, key=None, reverse=False):
        return IterChain(
            sorted(self._value, key=None if key is None else make_func(key), reverse=reverse)
        )

    def sum(self, start=UNSET):
        args = (start,) if start is not UNSET else ()
        return IterChain(sum(self._value, *args))

    if sys.version_info < (3, 10):

        def zip(self):
            return IterChain(zip(*self._value))

    else:

        def zip(self, strict=UNSET):
            kwargs = dict(strict=strict) if strict is not UNSET else {}
            return IterChain(zip(*self._value, **kwargs))

    ## itertools

    def starmap(self, f):
        return IterChain(starmap(f, self._value))

    def zip_longest(self, fillvalue=None):
        return IterChain(zip_longest(*self._value, fillvalue=fillvalue))

    ## heapq

    def nlargest(self, n, key=None):
        return IterChain(nlargest(n, self._value, key=None if key is None else make_func(key)))

    def nsmallest(self, n, key=None):
        return IterChain(nsmallest(n, self._value, key=None if key is None else make_func(key)))

    ## random

    def choice(self):
        return IterChain(choice(self._value))

    def choices(self, weights=None, *, cum_weights=None, k=1):
        return IterChain(choices(self._value, weights, cum_weights=cum_weights, k=k))

    if sys.version_info < (3, 9):

        def sample(self, k):
            return IterChain(sample(self._value, k))

    else:

        def sample(self, k, *, counts=None):
            return IterChain(sample(self._value, k, counts=counts))

    # The "random" argument to random.shuffle() was removed in Python 3.11.
    if sys.version_info < (3, 11):

        def shuffle(self, random=None):
            kwargs = dict(random=random) if random is not None else {}
            value = list(self._value)
            shuffle(value, **kwargs)
            return IterChain(value)

    else:

        def shuffle(self):
            value = list(self._value)
            shuffle(value)
            return IterChain(value)

    ## dicts

    def items(self):
        return IterChain(self._value.items())

    def values(self):
        return IterChain(self._value.values())

    def keys(self):
        return IterChain(self._value.keys())

    def update(self, *args, **kwargs):
        value = self._value.copy()
        value.update(*args, **kwargs)
        return IterChain(value)

    ## funcy.colls

    def join(self):
        return IterChain(funcy.join(self._value))

    merge = join

    def join_with(self, f):
        return IterChain(funcy.join_with(make_func(f), self._value))

    merge_with = join_with

    def walk(self, f):
        return IterChain(funcy.walk(make_func(f), self._value))

    def walk_keys(self, f):
        return IterChain(funcy.walk_keys(make_func(f), self._value))

    def walk_values(self, f):
        return IterChain(funcy.walk_values(make_func(f), self._value))

    def select(self, f):
        return IterChain(funcy.select(make_func(f), self._value))

    def select_keys(self, f):
        return IterChain(funcy.select_keys(make_func(f), self._value))

    def select_values(self, f):
        return IterChain(funcy.select_values(make_func(f), self._value))

    def compact(self):
        return IterChain(funcy.compact(self._value))

    def flip(self):
        return IterChain(funcy.flip(self._value))

    def project(self, keys):
        return IterChain(funcy.project(self._value, keys))

    def omit(self, keys):
        return IterChain(funcy.omit(self._value, keys))

    def zip_values(self):
        # funcy.zip_values() raises a TypeError for empty collections, but that's annoying.
        if len(self._value) == 0:
            return IterChain([])
        return IterChain(funcy.zip_values(*self._value))

    def zip_dicts(self):
        # funcy.zip_dicts() raises a TypeError for empty collections, but that's annoying.
        if len(self._value) == 0:
            return IterChain([])
        return IterChain(funcy.zip_dicts(*self._value))

    def where(self, **cond):
        return IterChain(funcy.where(self._value, **cond))

    def pluck(self, key):
        return IterChain(funcy.pluck(key, self._value))

    def pluck_attr(self, attr):
        return IterChain(funcy.pluck_attr(attr, self._value))

    def invoke(self, name, *args, **kwargs):
        return IterChain(funcy.invoke(self._value, name, *args, **kwargs))

    ## funcy.seqs

    def drop(self, n):
        return IterChain(funcy.drop(n, self._value))

    def take(self, n):
        return IterChain(funcy.take(n, self._value))

    def rest(self):
        return IterChain(funcy.rest(self._value))

    def butlast(self):
        return IterChain(funcy.butlast(self._value))

    def map(self, f):
        return IterChain(funcy.map(make_func(f), self._value))

    def filter(self, predicate):
        predicate = make_pred(predicate)
        return IterChain(funcy.filter(predicate, self._value))

    def remove(self, predicate):
        predicate = make_pred(predicate)
        return IterChain(funcy.remove(predicate, self._value))

    def keep(self, f=UNSET):
        if f is UNSET:
            return IterChain(funcy.keep(self._value))
        else:
            return IterChain(funcy.keep(make_func(f), self._value))

    def without(self, *items):
        return IterChain(funcy.without(self._value, *items))

    def concat(self):
        return IterChain(funcy.cat(self._value))

    def flatten(self, follow=UNSET):
        kwargs = dict(follow=follow) if follow is not UNSET else {}
        return IterChain(funcy.flatten(self._value, **kwargs))

    def mapcat(self, f):
        return IterChain(funcy.mapcat(make_func(f), self._value))

    def interleave(self):
        return IterChain(funcy.interleave(*self._value))

    def interpose(self, sep):
        return IterChain(funcy.interpose(sep, self._value))

    def takewhile(self, pred=UNSET):
        if pred is UNSET:
            return IterChain(funcy.takewhile(self._value))
        else:
            return IterChain(funcy.takewhile(pred, self._value))

    def dropwhile(self, pred=UNSET):
        if pred is UNSET:
            return IterChain(funcy.dropwhile(self._value))
        else:
            return IterChain(funcy.dropwhile(pred, self._value))

    def distinct(self, key=UNSET):
        kwargs = dict(key=key) if key is not UNSET else {}
        return IterChain(funcy.distinct(self._value, **kwargs))

    def split(self, pred):
        return IterChain(funcy.split(pred, self._value))

    def split_at(self, n):
        return IterChain(funcy.split_at(n, self._value))

    def split_by(self, pred):
        return IterChain(funcy.split_by(pred, self._value))

    def group_by(self, f):
        return IterChain(funcy.group_by(make_func(f), self._value))

    def group_by_keys(self, get_keys):
        return IterChain(funcy.group_by_keys(make_func(get_keys), self._value))

    def group_values(self):
        return IterChain(funcy.group_values(self._value))

    def count_by(self, f):
        return IterChain(funcy.count_by(make_func(f), self._value))

    def count_reps(self):
        return IterChain(funcy.count_reps(self._value))

    def partition(self, n, step=UNSET):
        if step is UNSET:
            return IterChain(funcy.partition(n, self._value))
        else:
            return IterChain(funcy.partition(n, step, self._value))

    def chunks(self, n, step=UNSET):
        if step is UNSET:
            return IterChain(funcy.chunks(n, self._value))
        else:
            return IterChain(funcy.chunks(n, step, self._value))

    def partition_by(self, f):
        return IterChain(funcy.partition_by(make_func(f), self._value))

    def with_prev(self, fill=None):
        return IterChain(funcy.with_prev(self._value, fill=fill))

    def with_next(self, fill=None):
        return IterChain(funcy.with_next(self._value, fill=fill))

    def pairwise(self):
        return IterChain(funcy.pairwise(self._value))

    def accumulate(self, func=UNSET):
        kwargs = dict(func=func) if func is not UNSET else {}
        return IterChain(funcy.accumulate(self._value, **kwargs))

    def reductions(self, f, acc=UNSET):
        kwargs = dict(acc=acc) if acc is not UNSET else {}
        return IterChain(funcy.reductions(make_func(f), self._value, **kwargs))

    def sums(self, acc=UNSET):
        kwargs = dict(acc=acc) if acc is not UNSET else {}
        return IterChain(funcy.sums(self._value, **kwargs))



import functools
import itertools
from collections.abc import Collection, Iterable, MappingView, Sequence

import pytest

from funcy_chain import Chain as _Chain
from funcy_chain import IterChain


class IterChainWrapper:
    """Wrapper for testing IterChain identically to Chain."""

    def __init__(self, value):
        self._iterchain = IterChain(value)

    @classmethod
    def from_iterchain(self, iterchain):
        return IterChainWrapper(iterchain._value)

    @property
    def value(self):
        value = self._iterchain.value
        if isinstance(value, Iterable) and (
            isinstance(value, MappingView) or not isinstance(value, Collection)
        ):
            value = list(value)
        return value

    def __getattr__(self, name):
        attr = getattr(self._iterchain, name)
        if callable(attr):

            @functools.wraps(attr)
            def new_method(*args, **kwargs):
                return IterChainWrapper.from_iterchain(attr(*args, **kwargs))

            return new_method
        return attr


@pytest.fixture(scope="session", params=[_Chain, IterChainWrapper])
def Chain(request):
    return request.param
